import os
import sys
import json
import copy
import shutil
import time
import imp
import tempfile
import pytest
import random
import string
import traceback
import moto
import boto3
from moto import mock_s3
import traceback
import pandas as pd

from enrichsdk.package.mock import MockConfig, MockState
from enrichsdk import Transform
from enrichsdk.core.mixins import S3Mixin, ParallelMixin, FilesMixin
from enrichsdk.lib import Context


class MixinInstance(Transform, S3Mixin, ParallelMixin, FilesMixin):
    pass


# => Smple dataset
carsfile = os.path.join(os.path.dirname(__file__), "fixtures", "data", "cars.csv")
carsdf = pd.read_csv(carsfile)


def cars_compute(rows):
    return rows.shape[0]


###########################
def id_generator(size=None, chars=string.ascii_uppercase + string.digits):
    if size is None:
        size = random.randint(1, 25)
    return "".join(random.choice(chars) for _ in range(size))


id_list = [id_generator() for i in range(100)]


def splitfunc(s):
    return len(s)


def applyfunc(label, strlist):
    return len(strlist)


test_bucket_name = "hello"
csv_files = {
    "txns/2014-01-01.csv": (
        b"name,amount,id\n" b"Alice,100,1\n" b"Bob,200,2\n" b"Charlie,300,3\n"
    ),
    "txns/2014-01-02.csv": (b"name,amount,id\n"),
    "txns/2015-01-03.csv": (
        b"name,amount,id\n" b"Dennis,400,4\n" b"Edith,500,5\n" b"Frank,600,6\n"
    ),
}

# Ideas on mocking s3 is from here
# https://github.com/dask/s3fs/blob/master/s3fs/tests/test_s3fs.py
port = 5555
endpoint_uri = "http://127.0.0.1:%s/" % port


def get_boto3_client():
    from botocore.session import Session

    # NB: we use the sync botocore client for setup
    session = Session()
    return session.create_client("s3", endpoint_url=endpoint_uri)


class TestClass(object):
    """
    Test whether I can instantiate a module
    """

    @classmethod
    def setup_class(cls):

        # => Boostrap the workspace
        enrich_root = tempfile.mkdtemp(prefix="enrichtest__")
        cls.enrich_root = enrich_root
        os.environ["ENRICH_ROOT"] = enrich_root

        # => Set root
        enrich_data = os.path.join(enrich_root, "data")
        try:
            os.makedirs(enrich_data)
        except:
            pass
        os.environ["ENRICH_DATA"] = enrich_data

        # => etc
        enrich_etc = os.path.join(enrich_root, "etc")
        try:
            os.makedirs(enrich_etc)
        except:
            pass
        os.environ["ENRICH_ETC"] = enrich_etc

        siteconf = {
            "credentials": {"aws": {"secret_key": "secret", "access_key": "access"}}
        }
        with open(os.path.join(enrich_etc, "siteconf.json"), "w") as fd:
            fd.write(json.dumps(siteconf, indent=4))

        context = Context()
        cls.config = MockConfig(context.asdict())

        # Make sure that site conf is present for the get_credentials.
        # to work
        cls.config.load_test_state({})

        cls.state = MockState(config=cls.config)
        cls.mixintest = MixinInstance(config=cls.config)

        # Need to run a s3 simulating server. See this:
        # https://github.com/dask/s3fs/blob/main/s3fs/tests/test_s3fs.py
        try:
            import shlex
            import subprocess

            cls.proc = subprocess.Popen(shlex.split("moto_server s3 -p %s" % port))
        except Exception as e:
            traceback.print_exc()

        os.environ["AWS_ACCESS_KEY_ID"] = "testing"
        os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
        os.environ["AWS_SECURITY_TOKEN"] = "testing"
        os.environ["AWS_SESSION_TOKEN"] = "testing"

        # Create s3 client
        cls.client = get_boto3_client()

        # Now
        cls.client.create_bucket(Bucket=test_bucket_name)

        # Put files
        for flist in [csv_files]:
            for f, data in flist.items():
                cls.client.put_object(
                    Bucket=test_bucket_name,
                    Key=f,
                    Body=data,
                    ContentType="application/csv",
                    ContentEncoding="utf-8",
                    StorageClass="STANDARD",
                    ACL="public-read-write",
                )

    @classmethod
    def teardown_class(cls):

        if os.path.exists(cls.enrich_root):
            shutil.rmtree(cls.enrich_root)

        for flist in [csv_files]:
            for f, data in flist.items():
                try:
                    cls.client.delete_object(Bucket=test_bucket_name, Key=f)
                except:
                    traceback.print_exc()
                    pass

        cls.proc.terminate()

    def test_invalid_handle_none(self):
        """
        Test invalid credentials
        """

        # Dont give any credentials...
        with pytest.raises(Exception) as excinfo:
            self.mixintest.get_s3_handle(client_kwargs={"endpoint_url": endpoint_uri})

        assert "Could not find the credentials" in str(excinfo)

    def test_invalid_handle_awscred(self):
        """
        Test invalid credentials via aws cred
        """

        self.mixintest.aws_cred = None

        # Dont give any credentials...
        with pytest.raises(Exception) as excinfo:
            self.mixintest.get_s3_handle(client_kwargs={"endpoint_url": endpoint_uri})

        assert "Could not find the credentials" in str(excinfo)

    def test_correct_handle(self):
        """
        Test s3_get_handle
        """

        cred = self.mixintest.get_credentials("aws")
        handle = self.mixintest.get_s3_handle(
            cred=cred, client_kwargs={"endpoint_url": endpoint_uri}
        )

        assert len(handle.ls("/")) == 1
        assert handle.ls("/")[0] == "hello"

    def test_list_files_invalid_s3(self):
        """
        Test list files with invalid s3
        """

        cred = self.mixintest.get_credentials("aws")
        s3 = None

        with pytest.raises(Exception) as excinfo:
            included, excluded = self.mixintest.s3_list_files(
                path="sss", bucket="hello", s3=s3
            )
        assert "Invalid s3 handle" in str(excinfo)

    def test_list_files_invalid_path(self):
        """
        Test list files with invalid path
        """

        cred = self.mixintest.get_credentials("aws")
        s3 = self.mixintest.get_s3_handle(
            cred=cred, client_kwargs={"endpoint_url": endpoint_uri}
        )

        included, excluded = self.mixintest.s3_list_files(
            path="sss", bucket="hello", s3=s3
        )

        assert len(included) == 0
        assert len(excluded) == 0

    def test_list_files_invalid_bucket(self):
        """
        Test list files with invalid path
        """

        cred = self.mixintest.get_credentials("aws")
        s3 = self.mixintest.get_s3_handle(
            cred=cred, client_kwargs={"endpoint_url": endpoint_uri}
        )

        with pytest.raises(Exception) as excinfo:
            included, excluded = self.mixintest.s3_list_files(path="sss", s3=s3)

        assert "Invalid/missing bucket" in str(excinfo)

    def test_list_files_valid_1(self):
        """
        Test list files with valid path - all files
        """

        cred = self.mixintest.get_credentials("aws")
        s3 = self.mixintest.get_s3_handle(
            cred=cred, client_kwargs={"endpoint_url": endpoint_uri}
        )

        included, excluded = self.mixintest.s3_list_files(
            path="txns/*", bucket="hello", s3=s3
        )

        assert len(included) == 3

    def test_list_files_valid_2(self):
        """
        Test list files with valid path - apply filter
        """

        cred = self.mixintest.get_credentials("aws")
        s3 = self.mixintest.get_s3_handle(
            cred=cred, client_kwargs={"endpoint_url": endpoint_uri}
        )

        included, excluded = self.mixintest.s3_list_files(
            path="txns/2015*", bucket="hello", s3=s3
        )

        assert len(included) == 1

    def test_list_files_invalid_func(self):
        """
        Test list files with invalid function
        """

        cred = self.mixintest.get_credentials("aws")
        s3 = self.mixintest.get_s3_handle(
            cred=cred, client_kwargs={"endpoint_url": endpoint_uri}
        )

        func = "hello"

        with pytest.raises(Exception) as excinfo:
            included, excluded = self.mixintest.s3_list_files(
                path="txns/*", bucket="hello", s3=s3, include=func
            )

        assert "Include" in str(excinfo)

    def test_list_files_valid_func(self):
        """
        Test list files with valid function
        """

        cred = self.mixintest.get_credentials("aws")
        s3 = self.mixintest.get_s3_handle(
            cred=cred, client_kwargs={"endpoint_url": endpoint_uri}
        )

        func = lambda x: x.startswith("txns/2014")

        included, excluded = self.mixintest.s3_list_files(
            path="txns/*", bucket="hello", s3=s3, include=func
        )

        assert len(included) == 2

    @pytest.mark.parametrize(
        "df,partitions,cores",
        [
            (carsdf, 1, 1),
            (carsdf, 2, 1),
            (carsdf, 2, 2),
            (carsdf, 2, 10),
            (carsdf, 5, 10),
        ],
    )
    def test_parallel_single(self, df, partitions, cores):
        mixintest = self.mixintest
        rows = df.shape[0]
        dfs = mixintest.pexec_single(df, cars_compute, partitions, cores)
        assert sum(dfs) == rows

    @pytest.mark.parametrize(
        "df,partitions,cores",
        [
            (carsdf, 1, 1),
            (carsdf, 2, 1),
            (carsdf, 2, 2),
            (carsdf, 2, 10),
            (carsdf, 5, 10),
        ],
    )
    def test_parallel_multiple(self, df, partitions, cores):
        """
        Test parallel execution
        """
        mixintest = self.mixintest

        dfs = [df[df.YEAR == y] for y in df.YEAR.unique()]
        dfs = mixintest.pexec_multiple(dfs, cars_compute, cores)
        assert sum(dfs) == df.shape[0]

    @pytest.mark.parametrize(
        "label,files,splitfunc,applyfunc,error",
        [
            (None, None, None, None, "label"),
            ("hello", None, None, None, "files"),
            ("hello", "Kilo", None, None, "files"),
            ("hello", [], None, None, "splitfunc"),
            ("hello", [], "shello", None, "splitfunc"),
            ("hello", [], splitfunc, None, "applyfunc"),
            ("hello", [], splitfunc, "hello", "applyfunc"),
        ],
    )
    def test_file_split_apply_invalid_parameters(
        self, label, files, splitfunc, applyfunc, error
    ):
        """
        Check parameters
        """

        with pytest.raises(Exception) as excinfo:
            self.mixintest.file_split_apply(label, files, splitfunc, applyfunc)

        assert "Invalid parameter" in str(excinfo)
        assert error in str(excinfo)

    def test_file_split_apply_valid_parameters(self):
        """
        Check valid parameters
        """

        result = self.mixintest.file_split_apply("ehllo", id_list, splitfunc, applyfunc)

        assert isinstance(result, dict)
        assert len(id_list) == sum(result.values())

    def test_file_preprocessed_read(self):
        """
        Test the file preprocess
        """

        # => Check whether the loading is correct
        root = os.path.join(self.enrich_root, "data", "_test", "root")
        metadata = self.mixintest.file_preprocessed_read(root)

        assert "dfs" in metadata
        assert "files" in metadata
        assert len(metadata["dfs"]) == 0
        assert len(metadata["files"]) == 0



"""
Get the environment object
"""
import os
import sys
import copy
import yaml
import json
import logging
import platform

logger = logging.getLogger()

class Context:
    def __init__(self, config=None, *args, **kwargs):

        self.context = {}

        if isinstance(config, dict) and len(config) > 0:
            config = None

        if config is None:
            if "ENRICH_CONTEXT" in os.environ:
                config = os.environ["ENRICH_CONTEXT"]
                if not os.path.exists(config):
                    raise Exception("Context file missing: {}".format(config))
            elif os.path.exists('context.yml'):
                config = "context.yml"
            elif os.path.exists('context.yaml'):
                config = "context.yaml"

        if isinstance(config, dict) and len(config) == 0:
            config = None

        if config is not None:
            if isinstance(config, str) and os.path.exists(config):
                try:
                    if config.endswith(".json"):
                        config = json.load(open(config))
                    elif (config.endswith(".yaml") or config.endswith(".yml")):

                        config = yaml.load(open(config), Loader=yaml.FullLoader)
                    else:
                        raise Exception("Unknown input format")
                except:
                    raise
                if isinstance(config, dict) and len(config) > 0:
                    self.context = config
            elif isinstance(config, dict) and len(config) > 0:
                self.context = config
            else:
                raise Exception("Invalid context file")

        # => Priority yaml config -> environment variables..
        # Root has to be specified
        if "ENRICH_ROOT" not in self.context:
            if "ENRICH_ROOT" in os.environ:
                self.context["ENRICH_ROOT"] = os.environ["ENRICH_ROOT"]
            else:
                logger.warning(
                    "ENRICH_ROOT must ideally be specified in context file or environment. Using default /tmp/enrich")
                self.context['ENRICH_ROOT'] = os.environ['ENRICH_ROOT'] = '/tmp/enrich'

        # WINDOWS PORTING
        opsys = platform.system()
        defaults = {
            "ENRICH_RUN_ROOT": "%(ENRICH_ROOT)s",
            "ENRICH_DATA": os.path.join("%(ENRICH_ROOT)s", "data"),
            "ENRICH_ETC": os.path.join("%(ENRICH_ROOT)s", "etc"),
            "ENRICH_SHARED": os.path.join("%(ENRICH_ROOT)s", "shared"),
            "ENRICH_VAR": os.path.join("%(ENRICH_ROOT)s", "var"),
            "ENRICH_LIB": os.path.join("%(ENRICH_ROOT)s", "lib"),
            "ENRICH_OPT": os.path.join("%(ENRICH_ROOT)s", "opt"),
            "ENRICH_LOGS": os.path.join("%(ENRICH_ROOT)s", "logs"),
            "ENRICH_CUSTOMERS": os.path.join("%(ENRICH_ROOT)s", "customers"),
            "ENRICH_RELEASES": os.path.join("%(ENRICH_ROOT)s", "releases"),
            "ENRICH_TEST": os.path.join("%(ENRICH_ROOT)s", "test"),
            "ENRICH_MANAGE": os.path.join("%(ENRICH_ROOT)s", "manage"),
            "versionmap": os.path.join("%(ENRICH_ROOT)s", "etc", "versionmap.json"),
            "django-env": os.path.join("%(ENRICH_ROOT)s", "etc", "django-env" if opsys != "Windows" else "django-env.bat"),
            "EMAIL_HOST": "unknown",
            "EMAIL_HOST_USER": "notset",
            "EMAIL_HOST_PASSWORD": "notset",
            "EMAIL_PORT": "9999",
            "EMAIL_USE_TLS": "TRUE",
        }

        for k, v in defaults.items():
            if k not in self.context:
                # If specified in the environment
                if k in os.environ:
                    self.context[k] = os.environ[k]
                else:
                    self.context[k] = v

        # For siteconf check the available alternatives. If none
        # present, then dont add the entry.
        k = "siteconf"
        if k not in self.context:
            if k in os.environ:
                self.context[k] = os.environ[k]
            else:
                arsiteconf = [
                    os.path.join("%(ENRICH_ROOT)s", "etc", "siteconf.jsonx"),
                    os.path.join("%(ENRICH_ROOT)s", "etc", "siteconf.json"),
                ]
                for s in arsiteconf:
                    try:
                        s = s % self.context
                        if os.path.exists(s):
                            self.context[k] = s
                            break
                    except:
                        pass

        # Resolve all of them...
        for k, v in self.context.items():
            if isinstance(v, str):
                v = v % self.context
                self.context[k] = v

        if "ENRICH_ROOT" not in self.context:
            print("Please define ENRICH_ROOT or provide a environmental context file")
            raise Exception("Missing ENRICH_ROOT. Cant build context")

    def set_var(self, k, v):
        """
        Update the value for given context variable
        """
        self.context[k] = v

    def get_var(self, k):
        """
        Get the value for given context variable
        """
        return self.context.get(k, None)

    def validate(self):

        for attr in ["ENRICH_ROOT", "ENRICH_DATA", "ENRICH_CUSTOMERS", "ENRICH_ETC"]:
            if attr not in self.context:
                raise Exception("Incomplete context")

    def set_env(self):
        """
        Set the environment using the context
        """
        update = {k: v for k, v in self.context.items() if k.isupper()}
        os.environ.update(update)

    def asdict(self):
        return copy.copy(self.context)



"""
Simple LLM Proxy that works with the front end. This derives from the QnA service module.

"""
import os
import sys
import json
import traceback
from datetime import datetime

from logging.config import dictConfig
from typing import Annotated, Dict, Any, Optional, List
from enum import Enum

from pydantic import BaseModel,  ValidationError, validator
from fastapi import FastAPI, File, Form, UploadFile
from fastapi import HTTPException, Header, Security
from fastapi.security import APIKeyHeader, APIKeyCookie, APIKeyQuery

from llmsdk.lib import SafeEncoder
from llmsdk.services.log import *
from llmsdk.services.lib import get_stats, initialize_stats

# Override default uvicorn logger...
logging.config.dictConfig(log_config)
logger = get_logger()

HEADER_API_KEY = APIKeyHeader(name='X-API-Key',auto_error=False)
def validate_api_key(key):
    """
    Check whether the API Key is valid
    """
    api_keys = os.environ.get("APIKEYS","")
    api_keys = api_keys.split(",")

    return key in api_keys

app = FastAPI()

class FormatChoices(Enum):
    xml = 'xml'
    pdf = 'pdf'
    json = 'json'

class CategoryChoices(Enum):
    test = 'test'
    production = 'production'

class Document(BaseModel):
    uuid: Optional[str] = None
    filename: str
    content: str
    metadata: Dict[Any,Any] = {}


class ListenerData(BaseModel):
    category: CategoryChoices
    uuid: str
    documents: List[Document]
    metadata: Dict[Any,Any] = {}

    @validator('uuid')
    def uuid_must_be_10chars_long(cls, v):
        if len(v) < 10:
            raise ValueError('must be atleast 10 characters long')
        return v

    @validator('documents')
    def atleast_one_document(cls, v):
        if len(v) == 0:
            raise ValueError('must include atleast one document')
        return v

@app.get("/health")
def health():
    """
    Return usage statistics

    Returns
    -------
    stats: dict
           Usage statistics

    """

    stats = get_stats()

    logger.info("Returning stats",
                extra={
                    'data': json.dumps(stats, indent=4)
                })
    return stats


@app.post("/application/")
async def receive_application(
        data: ListenerData,
        header_api_key: str = Security(HEADER_API_KEY),
):

    """
    Post a new application and receive an acknowledgement

    Returns
    -------
    stats: dict
           uuid, size, and type of the application
    """

    if not validate_api_key(header_api_key):
        logger.error("Invalid API Key",
                     extra={
                         'data': header_api_key
                     })
        raise HTTPException(status_code=401,
                            detail=f"Invalid API Key")

    # Do a check on API key
    # Now check if the document is new or existing...
    data_root = os.environ['DATA_ROOT']

    doc_id = data.uuid
    doc_category = data.category.value
    doc_metadata = data.metadata

    targetdir = os.path.join(data_root, doc_category, doc_id)
    os.makedirs(targetdir, exist_ok=True)

    metadatafile = os.path.join(targetdir, "metadata.json")
    successfile = os.path.join(targetdir, "_SUCCESS")
    if ((os.path.exists(targetdir)) and
        (os.path.exists(metadatafile)) and
        (os.path.exists(successfile))):
        logger.error("Document exists",
                     extra={
                         'request_id': doc_id,
                     })
        raise HTTPException(status_code=403,
                            detail= f"Document already uploaded: {doc_id}")

    try:

        metadata = {
            'utctimestamp': datetime.utcnow().isoformat(),
            "dirname": targetdir,
            "document": {
                "id": doc_id,
                "metadata": doc_metadata,
                "files": {}
            }
        }

        # Dump the files...
        for doc in data.documents:
            doc_content = doc.content
            doc_filename = doc.filename
            doc_metadata = doc.metadata

            filename = os.path.join(targetdir, doc_filename)
            with open(filename, "w") as fd:
                fd.write(doc_content)

            # Store the metadata
            metadata['document']['files'][doc_filename] = {
                "size": len(doc_content),
                "metadata": doc_metadata
            }

        # Write the metadata
        with open(metadatafile, 'w') as fd:
            fd.write(json.dumps(metadata, indent=4))

        with open(successfile, "w") as fd:
            fd.write("")

        logger.info("Document received",
                    extra={
                        'request_id': doc_id,
                        'data': json.dumps(metadata, indent=4)
                    })

        return {
            "doc_id": doc_id,
            "status": "success",
            "message": "Received document",
        }

    except:
        traceback.print_exc()
        logger.exception("Error while process",
                         extra={
                             'request_id': doc_id,
                             'data': targetdir
                         })

        raise HTTPException(status_code=500,
                            detail= f"Error while processing the document")

@app.on_event("startup")
def app_startup():


    if "DATA_ROOT" not in os.environ:
        print("DATA_ROOT is missing in the environment")
        raise Exception("Invalid configuration")

    initialize_stats()



import os
import sys
import json
import requests
import logging
import time
import csv
import re
import traceback

from django.urls import reverse, resolve
from django.conf.urls import url, include
from django.contrib import messages
from django.shortcuts import render
from django.http import HttpResponseRedirect, Http404, JsonResponse, HttpResponse
from django.utils.timezone import localtime

from dateutil import parser as dateparser

from enrichsdk.app.utils import clean_and_validate_widgets
from enrichsdk.lib import get_credentials_by_name

logger = logging.getLogger("app")

class SalesforceClient:

    def __init__(self, cred):

        if isinstance(cred, str):
            self.cred = get_credentials_by_name(cred)
        else:
            self.cred = cred
        self.baseurl = self.cred['url']
        self.token = None
        self.salesforce_version = "v58.0"

    def get_token(self, force=False):
        """
        Get access token...
        """
        sample_token = {
            'access_token': '00D7j000000H9Ht!AR...',
            'instance_url': 'https://acmecompanycompany--preprod.sandbox.my.salesforce.com',
            'id': 'https://test.salesforce.com/id/00D7j000000H9HtEAK/0057j0000053bRfAAI',
            'token_type': 'Bearer',
            'issued_at': '1697972695996',
            'signature': 'BQ+dEwXSrqZcZtqXYGSYR2B+9+3eftIeBjT92Dv2YYI='
        }

        # Assumption. We dont know when the token will expire. It is
        # said to be 2 hours in documentation
        timeout = 3600*1000
        now = int(1000*time.time())
        if ((not force) and
            (self.token is not None) and
            (isinstance(self.token, dict)) and
            ('access_token' in self.token) and
            (now < (int(self.token['issued_at']) + timeout))):
            return self.token['access_token']

        tokenurl = self.baseurl + "/services/oauth2/token"
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        # Construct the oauth request
        cred = self.cred
        data = {
            "grant_type": "password",
            "client_id": cred['client_id'],
            "client_secret": cred['client_secret'],
            "username": cred['username'],
            "password": cred['password'] + cred['client_security_token']
        }

        msg = ""
        msg += f"Token URL: {tokenurl}\n"
        for k, v in data.items():
            if k != "password":
                msg += f"{k}: {str(v)[:8]}...\n"
            else:
                msg += f"{k}: ****...\n"
            result = requests.post(tokenurl, data=data, headers=headers)
        try:
            self.token = result.json()
            if result.status_code == 200:
                logger.debug("Salesforce token obtained",
                             extra={
                                 'data': msg
                             })
            else:
                logger.error("Failed to obtained Salesforce token",
                             extra={
                             'data': msg + str(result.content)
                             })
        except:
                logger.exception("Failed to obtained Salesforce token",
                             extra={
                             'data': msg + str(result.content)
                             })

        return self.token['access_token']

    def access_salesforce(self, url, method="get",
                          params={}, data={},
                          request=None):

        token = self.get_token()
        url = self.baseurl + url
        headers = {
            'Authorization': f"Bearer {token}"
        }

        if method == "get":
            result = requests.get(url, params=params, headers=headers)
        elif method == "post":
            result = requests.post(url, params=params, headers=headers, json=data)
        elif method == "patch":
            result = requests.patch(url, params=params, headers=headers, json=data)
        else:
            raise Exception(f"Unknown access method: {method}")

        if result.status_code >= 400:
            logger.error("Failed to access Salesforce",
                         extra={

                             'data': f"URL: {url}\nOutput: {result.content}"
                         })
        try:
            if method != "patch":
                status, result = result.status_code, result.json()
            else:
                status, result = result.status_code, {}

            # [{"message":"Jurisdiction: bad value for restricted picklist field: State of Washington","errorCode":"INVALID_OR_NULL_FOR_RESTRICTED_PICKLIST","fields":["Jurisdiction__c"]}]
            if ((request is not None) and
                (isinstance(result, (dict, list))) and
                (len(result) > 0)):
                res = result
                if isinstance(res, list):
                    res = res[0]
                if isinstance(res, dict) and ("message" in res):
                    messages.error(request, "Salesforce message: " + res['message'])

            if 'nextRecordsUrl' in result:
                messages.error(request, "Internal error. A few search results were not processed. Please contact support")

            return status, result
        except:
            logger.exception("Failed to access Salesforce",
                             extra={
                                 'data': f"URL: {url}\n"
                             })

        raise Exception("Failed to access Salesforce")

    def run_query(self, query, request=None):

        query = re.split(r"\s+", query)
        query = "+".join(query)
        opurl = f"/services/data/{self.salesforce_version}/query/?q={query}"

        status, result = self.access_salesforce(opurl, request=request)
        if status >= 400:
            raise Exception("Failed to run query")
        return result

    def get_opportunity_by_id(self, oppid, request=None):

        opurl = f"/services/data/{self.salesforce_version}/sobjects/Opportunity/{oppid}"
        status, result = self.access_salesforce(opurl, request=request)
        if status >= 400:
            raise Exception("Failed to get opportunity")
        return result

    def describe_opportunity(self, request=None):

        opurl = f"/services/data/{self.salesforce_version}/sobjects/Opportunity/describe"
        status, result = self.access_salesforce(opurl, request=request)
        if status >= 400:
            raise Exception("Failed to describe opportunity")
        return result

    def get_opportunities(self, limit=200, columns=None, request=None):

        fields = "FIELDS(ALL)"
        if columns is not None:
            fields = ",".join(columns)

        opurl = f"/services/data/{self.salesforce_version}/query/?q=SELECT+{fields}+FROM+Opportunity+ORDER+BY+LastModifiedDate+DESC+LIMIT+{limit}"

        status, result = self.access_salesforce(opurl, request=request)
        if status >= 400:
            raise Exception("Failed to query for opportunities")
        return result

    def get_opportunity_detail(self, oppid, request=None):

        opurl = f"/services/data/{self.salesforce_version}/sobjects/Opportunity/{oppid}"

        status, result = self.access_salesforce(opurl, request=request)
        if status >= 400:
            raise Exception("Failed to get opportunity detail")
        return result

    def add_opportunity(self, data, request=None):
        opurl = f"/services/data/{self.salesforce_version}/sobjects/Opportunity"
        status, result = self.access_salesforce(opurl, method="post", data=data, request=request)
        if status >= 400:
            raise Exception("Failed to add opportunity")
        return result

    def update_opportunity(self, oppid, data, request=None):
        opurl = f"/services/data/{self.salesforce_version}/sobjects/Opportunity/{oppid}"
        status, result = self.access_salesforce(opurl,
                                                method="patch",
                                                data=data,
                                                request=request)
        if status >= 400:
            raise Exception("Failed to add opportunity")
        return result

    def get_accounts(self, limit=200, columns=None, request=None):

        fields = "FIELDS(ALL)"
        if columns is not None:
            fields = ",".join(columns)

        opurl = f"/services/data/{self.salesforce_version}/query/?q=SELECT+{fields}+FROM+Account+ORDER+BY+LastModifiedDate+DESC+LIMIT+{limit}"

        status, result = self.access_salesforce(opurl, request=request)
        if status >= 400:
            raise Exception("Failed to query for opportunities")
        return result

    def add_account(self, data, request=None):
        opurl = f"/services/data/{self.salesforce_version}/sobjects/Account"
        status, result = self.access_salesforce(opurl,
                                                method="post",
                                                data=data, request=request)
        if status >= 400:
            raise Exception("Failed to add account")
        return result

    def get_account_by_id(self, accid, request=None):

        opurl = f"/services/data/{self.salesforce_version}/sobjects/Account/{accid}"
        status, result = self.access_salesforce(opurl, request=request)
        if status >= 400:
            raise Exception(f"Failed to get account details: {accid}")
        return result

#############################################
#=> Fields in Salesforce Opportunity Object
#############################################
# Account_Name__c: null,
# Account_Region__c: null,
# Active_Headcount__c: null,
# Amount: 4810308.0,
# Amount_Lost__c: 0.0,
# Amount_M__c: 5.0,
# LastActivityDate: null,
# LastAmountChangedHistoryId: null,
# LastCloseDateChangedHistoryId: null,
# LastModifiedById: 0057j0000053bRfAAI,
# LastModifiedDate: 2023-10-19T17:37:53.000+0000,
# ....

class SalesforceBaseMixin:

    def salesforce_update_urlpatterns(self, prefix, urlpatterns):
        urlpatterns.extend([
            url(f'^{prefix}[/]?$', self.salesforce_index, name="salesforce_index"),
            url(f'^{prefix}/detail/(?P<oppid>[a-zA-Z0-9]+)[/]?$', self.salesforce_detail, name="salesforce_detail"),
        ])

    def salesforce_update_templates(self, templates):
        templates.update({
            'salesforce_index': 'sharedapp/generic_index.html',
            'salesforce_detail': 'sharedapp/generic_index.html',
        })

    def salesforce_get_server_name(self, request, spec):
        return "Salesforce"

    ######################################################
    # Matching
    ######################################################
    def get_client(self, request, spec):

        # Get the salesforce client...
        cred = spec['cred']
        cred = get_credentials_by_name(cred)
        if hasattr(self, "salesforce_client"):
            client = self.salesforce_client
        else:
            client = self.salesforce_client = SalesforceClient(cred)

        return cred, client

    def find_overlap_str(self, new, existing):

        if not isinstance(existing, str) or len(existing) == 0:
            return 0

        skip = ['pension', 'plan', 'inc', 'retirement']

        new = re.split(r"[\s\.,]+", new.lower().strip())
        new = [n for n in new if len(n) >= 3 and (n not in skip)]
        if len(new) == 0:
            return 0

        existing = re.split(r"[\s\.,]+", existing.lower().strip())
        existing = [n for n in existing if len(n) >= 3 and (n not in skip)]
        if len(existing) == 0:
            return 0

        overlap = len([x for x in new if x in existing])/len(existing)

        return overlap

    def find_overlap_name(self, new, existing):
        overlap = self.find_overlap_str(new, existing)
        overlap = int(overlap * 100)
        return overlap

    def find_overlap_date(self, dt1, dt2, days=30):

        try:
            if not isinstance(dt1, (date, datetime)):
                dt1 = dateparser.parse(dt1)
        except:
            return 0

        try:
            if not isinstance(dt2, (date, datetime)):
                dt2 = dateparser.parse(dt2)
        except:
            return 0

        diff = dt2-dt1
        return 1 if abs(diff.days) < days else 0

    def find_overlap_numeric(self, amt1, amt2, threshold=2):

        if ((amt1 is None) or
            (pd.isnull(amt1)) or
            (amt2 is None) or
            (pd.isnull(amt2)) or
            (not isinstance(amt1, (int, float))) or
            (not isinstance(amt2, (int, float))) or
            (amt1 <= 0) or
            (amt2 <= 0)):
            return 0

        ratio = (amt1/amt2) if amt1 > amt2 else (amt2/amt1)
        return 1 if ratio < threshold else 0

    def find_overlap_combined(self, modelvars, opportunity):

        total_overlap = 0

        for col in ['Plan Name', 'State']:
            overlap = self.find_overlap_name(modelvars[col], opportunity[col])
            total_overlap += 1

        return int((100*total_overlap/4))


    #################################################################
    # All Salesforce
    #################################################################
    def salesforce_index_get_extra_header_components(self, request, spec):

        r = resolve(request.path)
        return []

    def salesforce_index_finalize_entry(self, request, spec, opportunity, entry):
        return entry

    def salesforce_index_finalize_widgets(self, request, spec,
                                          opportunities, widgets):
        return widgets

    def salesforce_index_get_extra_actions(self, request, spec,
                                           opportunity,
                                           entry):
        return {}, []

    def salesforce_index_get_opportunities(self, request, spec):

        cred, client = self.get_client(request, spec)

        opportunities = client.get_opportunities(columns=[
            "Name",
            "Amount",
            "Id",
            "CreatedDate"
        ], request=request)

        return opportunities

    def salesforce_index_get_columns(self, request, spec, data):

        columns = [
            "Added", 'Name', "Amount (M$)",
        ]

        workflowcols = []
        detailcols = []

        if len(data) > 0:
            for k in data[0].keys():
                if k in columns:
                    continue
                if k.startswith("ACTION_"):
                    detailcols.append(k)
                else:
                    workflowcols.append(k)
        columns += [
            ('Workflow', workflowcols),
            ('Details', detailcols)
        ]
        return columns

    def salesforce_index_finalize_data(self, request, spec, data):
        return data

    def salesforce_index(self, request, spec):

        r = resolve(request.path)

        usecase = spec['usecase']
        namespace = spec['namespace']
        cred = spec['cred']
        if isinstance(cred, str):
            cred = get_credentials_by_name(spec['cred'])

        # First get the opportunities
        opportunities = self.salesforce_index_get_opportunities(request, spec)

        workflowcols = []
        data = []
        for o in opportunities.get('records',[]):

            amount = o['Amount']
            if amount is None:
                amount = 0
            amount = round(amount/10**6, 1)

            dt = dateparser.parse(o['CreatedDate'])
            detailurl = reverse(r.namespace + ":salesforce_detail",
                                kwargs={
                                    'oppid': o['Id']
                                })
            entry = {
                "Added": dt.replace(microsecond=0).strftime("%Y-%m-%d"),
                "Amount (M$)": amount,
                "Name": f'<a href="{detailurl}">{o["Name"]}</a>',
                "ACTION_SALESFORCE": {
                    "title": "Details",
                    "alt": "",
                    "class": "",
                    "template": "action_icon_compact",
                    "target": "_blank",
                    "icon": "salesforce_24x24",
                    "url": f"{cred['url']}/{o['Id']}"
                },
            }

            # Any cleanup before adding extra actions...
            entry = self.salesforce_index_finalize_entry(request, spec, o, entry)

            # Now add actions...
            extra, order = self.salesforce_index_get_extra_actions(request, spec, o, entry)
            entry.update(extra)

            data.append(entry)


        # How should I structure the output columns
        columns = self.salesforce_index_get_columns(request, spec, data)

        # Any header actions..
        extra_header_components = self.salesforce_index_get_extra_header_components(request, spec)

        server = self.salesforce_get_server_name(request, spec)
        widget = {
            "name": f"Opportunities in {server}",
            "description": f"Recent entries and not the complete list",
            "type": "full_width_table_compact_actions",
            "columns": columns,
            "search": True,
            "rows": data,
            "order": [[0, "desc"]],
            "td_class": "white-space-normal wordwrap",
            "thead_th_class": "",
            "header_components": {
                "components": [
                    {
                        "template": "action_search"
                    }
                ]
            }
        }

        widgets = [widget]

        # Do any extra cleanup
        widgets = self.salesforce_index_finalize_widgets(request, spec,
                                                         opportunities,
                                                         widgets)

        clean_and_validate_widgets(widgets)

        server = self.salesforce_get_server_name(request, spec)
        data = {
            "title": server,
            "sidebar_targets": self.get_sidebar(request, spec),
            "breadcrumb": server,
            "widgets": widgets
        }

        # Cleanup and add any final note..
        data = self.salesforce_index_finalize_data(request, spec, data)

        template = self.get_template(spec, 'salesforce_index')
        return render(request,
                      template,
                      {
                          'app': self,
                          'usecase': usecase,
                          'spec': spec,
                          'basenamespace': r.namespace,
                          'data': data
                      })

    def salesforce_detail_get_opportunity(self, request, spec, oppid):

        cred, client = self.get_client(request, spec)

        return client.get_opportunity_by_id(oppid, request=request)

    def salesforce_detail(self, request, spec, oppid):

        r = resolve(request.path)
        server = self.salesforce_get_server_name(request, spec)

        usecase = spec['usecase']
        namespace = spec['namespace']

        sfindexurl = reverse(r.namespace + ":salesforce_index")

        widgetspecs = []

        # => Get the opportunity object...
        try:
            detail = self.salesforce_detail_get_opportunity(request, spec, oppid)
        except:
            traceback.print_exc()
            logger.exception(f"Invalid opportunity: {oppid}")
            messages.error(request, f"Invalid opportunity: {oppid}")
            return HttpResponseRedirect(sfindexurl)

        accounts = {}
        data = []
        for name, value in detail.items():
            if not isinstance(value, str):
                value = str(value)
            data.append({
                "Attribute": name,
                "Value": value
            })

        name = detail['Name']
        widgetspecs.append({
            "name": name,
            "description": f"Opportunity detail in {server}",
            "data": data
        })

        columns = [
            "Attribute", "Value"
        ]

        widgets = []
        for widgetspec in widgetspecs:
            widget = {
                "name": widgetspec['name'],
                "description": widgetspec['description'],
                "type": "full_width_table_compact_actions",
                "columns": columns,
                "search": True,
                "rows": widgetspec['data'],
                "order": [[0, "asc"]],
                "td_class": "white-space-normal wordwrap",
                "thead_th_class": "",
                "header_components": {
                    "components": [
                        {
                            "template": "action_search"
                        }
                    ]
                }
            }
            widgets.append(widget)

        clean_and_validate_widgets(widgets)

        server = self.salesforce_get_server_name(request, spec)
        data = {
            "title": server,
            "sidebar_targets": self.get_sidebar(request, spec),
            "breadcrumbs": [
                {
                    "name": server,
                    "url": reverse(r.namespace + ":salesforce_index"),
                },
                {
                    "name": "Detail"
                }
            ],
            "widgets": widgets
        }

        template = self.get_template(spec, 'salesforce_detail')
        return render(request,
                      template,
                      {
                          'app': self,
                          'usecase': usecase,
                          'spec': spec,
                          'basenamespace': r.namespace,
                          'data': data
                      })


class SalesforceDatasetMixin:
    """
    Code had handles the integration between salesforce and Dataset
    """
    def salesforce_dataset_update_urlpatterns(self, prefix, urlpatterns):

        urlpatterns.extend([
            url(f'^{prefix}/post/(?P<datasetpk>[0-9a-zA-Z-_ .]+)/action[/]?$', self.salesforce_dataset_action, name="salesforce_dataset_action"),
            url(f'^{prefix}/post/(?P<datasetpk>[0-9a-zA-Z-_ .]+)/select[/]?$', self.salesforce_dataset_select, name="salesforce_dataset_select"),
            url(f'^{prefix}/post/(?P<datasetpk>[0-9a-zA-Z-_ .]+)/verify[/]?$', self.salesforce_dataset_verify, name="salesforce_dataset_verify"),
        ])

    def salesforce_dataset_update_templates(self, templates):
        templates.update({
            'salesforce_dataset_action': 'sharedapp/generic_index.html',
            'salesforce_dataset_verify': 'sharedapp/generic_index.html',
            'salesforce_dataset_select': 'sharedapp/generic_index.html'
        })

    def salesforce_dataset_select(self, request, spec, datasetpk):
        """
        Select whether to create or update
        """

        testmode = spec.get('testmode', False)
        server = self.salesforce_get_server_name(request, spec)

        usecase = spec['usecase']
        namespace = spec['namespace']
        r = resolve(request.path)

        cred, client = self.get_client(request, spec)

        dataseturl = reverse(r.namespace + ":dataset_index")
        validateurl = reverse(r.namespace + ":dataset_validate",
                              kwargs={
                                  'pk': datasetpk
                              })
        verifyurl = reverse(r.namespace + ":salesforce_dataset_verify",
                        kwargs={
                                'datasetpk': datasetpk
                        })
        redirect = HttpResponseRedirect(validateurl)

        # => Get models
        LLMDatasetModel = self.get_model(spec, 'llmdataset')

        # Get the dataset
        try:
            dataset = LLMDatasetModel.objects.get(pk=datasetpk)
        except LLMDatasetModel.DoesNotExist:
            messages.error(request, 'Invalid dataset')
            return HttpResponseRedirect(reverse(r.namespace + ":dataset_index"))

        ##########################################################
        # => First get all the model parameters
        ##########################################################
        try:
            modelvars = dataset.values.get('modelvar', [])
            if len(modelvars) == 0:
                messages.error(request, "Not posted. Please save first")
                return redirect

            # Convery into a dictionary
            modelvars = { v['name'] : v['value'] for v in modelvars}

        except:
            logger.exception("Unable to post to salesforce")
            messages.error(request, f'Internal error. Unable to post to {server}')
            return redirect

        ##########################################################
        #=> Extract Some basic information...
        ##########################################################
        plan_name = modelvars['Plan Name']
        plan_sponsor = modelvars['Plan Sponsor']
        consultant = modelvars['Consultant']
        estimated_size = modelvars['Estimated Size']
        state = modelvars['State']

        ##########################################################
        #=> Now search last N opportunities
        ##########################################################
        opportunities = client.get_opportunities(columns=[
            "Name",
            "Plan Name",
            "Plan Sponsor",
            "CreatedDate",
            "Due Date",
            "Estimated Size",
            "Deal Type",
            "Amount",
            "Stage",
            "Consultant",
            "Id"
        ], request=request)
        data = []
        for o in opportunities.get('records',[]):

            name = o["Plan Name"]

            # Keep only test. For dev/testing...
            if testmode:
                 if ((name is None) or ("ABC" not in name)):
                     continue

            dt = dateparser.parse(o['CreatedDate'])
            amount = o['Amount']
            if amount is None:
                amount = 0
            if amount > 100000:
                amount = round(amount/10**6, 1)

            detailurl = reverse(r.namespace + ":salesforce_detail",
                                kwargs={
                                    'oppid': o['Id']
                                })

            overlap = self.find_overlap_combined(modelvars, o)

            data.append({
                "Conflict Score": overlap,
                "Added": localtime(dt.replace(microsecond=0)),
                "Name": f'<a href="{detailurl}">{o["Name"]}</a>',
                "Plan Sponsor": o['Plan Sponsor'],
                "Consultant": o['Consultant'],
                "Deal Type": o["Deal Type"],
                "State": o["State"],
                "Amount (M$)": amount,
                "ACTION_EDIT": {
                    "title": "Edit This Opportunity",
                    "alt": "",
                    "class": "btn",
                    "template": "action_icon_compact",
                    "icon": "edit_24dp_1",
                    "text": "Update",
                    "url": verifyurl + f"?action=edit&opportunityid={o['Id']}"
                },
            })

        columns = [
            "Conflict Score", "Added", 'Name', "Plan Sponsor",
            "Consultant","Deal Type", "State",
            "Amount (M$)",
            ("Action", ["ACTION_EDIT"])
        ]

        widget = {
            "name": f"Update Existing Opportunity: {plan_name}",
            "description": "Select opportunity to edit. Note that this shows ONLY the most recent 75 opportunities. Make sure you look for placeholder opportunities as well.",
            "type": "full_width_table_compact_actions",
            "columns": columns,
            "search": True,
            "rows": data,
            "order": [[0, "desc"], [1, "desc"]],
            "td_class": "white-space-normal wordwrap",
            "thead_th_class": "",
            "header_components": {
                "components": [
                    {
                        "template": "action_search"
                    }
                ]
            }
        }

        widgets = [widget]

        #=> Add notes..
        widget = {
            "name": f"Create Opportunity: {plan_name}",
            "description": "Notes on the post",
            "type": "full_width_form",
            "submit": "Create",
            "action": verifyurl,
            "hidden_vars": {
                "action": "create"
            },
            "text": f"""\
<ol>
<li>Plan Name: {plan_name}</li>
<li>Plan Sponsor: {plan_sponsor}</li>
<li>Consultant: {consultant}</li>
<li>State: {state}</li>
<li>Estimated Size: {estimated_size}</li>
</ol>
<p>Note that state, estimated size and other variables will be cleaned to make sure it is consistent with database format. Make sure you check the "New Value" column for all variables</p>
""",
            "td_class": "white-space-normal wordwrap",
            "header_components": {
                "components": [
                    {
                        "template": "action_icon_compact",
                        "icon": "add_circle_outline_24dp_1",
                        "url": verifyurl + f"?action=create",
                        "title": "Create new opportunity",
                    },
                ]
            }
        }
        widgets.insert(0, widget)

        widget = {
            "name": "RFP Workflow",
            "type": "workflow",
            "order": [
                {
                    "name": "Upload RFP",
                    "url": dataseturl
                },
                {
                    "name": "Validate Parameters",
                    "url": validateurl
                },
                f"Prepare Post to {server}",
                f"Approve Post to {server}",
            ],
            "currpos": 2,
            "class": "mb-2"
        }
        widgets.insert(0, widget)

        clean_and_validate_widgets(widgets)

        data = {
            "title": f"Update or Create Opportunity: {dataset.name}",
            "sidebar_targets": self.get_sidebar(request, spec),
            "breadcrumbs": [
                {
                    "name": "Datasets",
                    "url": reverse(r.namespace + ":dataset_index")
                },
                {
                    "name": "Validate",
                    "url": validateurl
                },
                {
                    "name": f"Approve {server} Post"
                }
            ],
            "widgets": widgets
        }

        template = self.get_template(spec, 'salesforce_dataset_select')
        return render(request,
                      template,
                      {
                          'app': self,
                          'usecase': usecase,
                          'spec': spec,
                          'basenamespace': r.namespace,
                          'data': data,
                      })

    def field_specific_cleanup(self, field, value):

        # Floating point number
        if field in ['Estimated Size']:
            matches = re.findall(r"(\d+\.\d+|\d+)]?", value.replace(",",""))

            if matches is None:
                value = None
            else:
                value = float(matches[0])
                if value > 10**6:
                    # exact number is mentioned.
                    value = round(value/10**6, 2)

        if field in [
                "Final Quote",
                "Request for Quotation",
                "Intent Due",
                "Prelim Quotes Due",
                "Final Quote",
                "Premium Transfer Due",
                "Liability Assumption Date",
                "First Payment (Financial)"
        ]:
            try:
                value = dateparser.parse(value).date().isoformat()
            except:
                # traceback.print_exc()
                value = None

        return value

    def salesforce_dataset_verify(self, request, spec, datasetpk):
        """
        Post to salesforce
        """

        server = self.salesforce_get_server_name(request, spec)
        usecase = spec['usecase']
        namespace = spec['namespace']
        r = resolve(request.path)

        action = request.GET.get('action', None)
        action_oppid = request.GET.get('opportunityid', None)
        action_oppid = None if str(action_oppid).lower() == "none" else action_oppid

        dataseturl = reverse(r.namespace + ":dataset_index")
        validateurl = reverse(r.namespace + ":dataset_validate",
                              kwargs={
                                  'pk': datasetpk
                              })
        redirect = HttpResponseRedirect(validateurl)
        selecturl = reverse(r.namespace + ":salesforce_dataset_select",
                              kwargs={
                                  'datasetpk': datasetpk
                              })
        selectredirect = HttpResponseRedirect(selecturl)

        # => Is this an edit or create...
        if action not in ['create', 'edit']:
            messages.error(request, f"Invalid {server} action. Should be create or edit. Not {action}")
            return selectredirect

        if ((action == "edit") and
            ((not isinstance(action_oppid, str)) or
             (len(action_oppid) <= 16))):
            messages.error(request, f"Invalid {server} opportunity ID. Should have atleast 16 characters {len(action_oppid)}")
            return selectredirect

        cred, client = self.get_client(request, spec)

        # => Get models
        LLMDatasetModel = self.get_model(spec, 'llmdataset')

        # Get the dataset
        try:
            dataset = LLMDatasetModel.objects.get(pk=datasetpk)
        except LLMDatasetModel.DoesNotExist:
            messages.error(request, 'Invalid dataset')
            return HttpResponseRedirect(reverse(r.namespace + ":dataset_index"))

        ##########################################################
        # => First get all the model parameters
        ##########################################################
        try:
            modelvars = dataset.values.get('modelvar', [])
            if len(modelvars) == 0:
                messages.error(request, "Not posted. Please save first")
                return redirect

            # Convery into a dictionary
            varposition = { v['name'] : f"{idx:02d}" for idx, v in enumerate(modelvars)}
            modelvars = { v['name'] : v['value'] for v in modelvars}

        except:
            logger.exception("Unable to post to salesforce")
            messages.error(request, f'Internal error. Unable to post to {server}')
            return redirect

        ##########################################################
        #=> Some basic information...
        ##########################################################
        plan_name = modelvars['Plan Name']
        plan_sponsor = modelvars['Plan Sponsor']
        consultant = modelvars['Consultant']

        ##########################################################
        # => Existing opportunity
        ##########################################################
        opportunity = {}
        if action == "edit":
            try:
                opportunity = client.get_opportunity_by_id(action_oppid, request=request)
            except:
                messages.error(request, f"Invalid opportunity ID: {action_oppid}")
                logger.exception(f"Invalid opportunity ID: {action_oppid}")
                return selectredirect

        ##########################################################
        # => Now construct the widgets...
        ##########################################################
        def detect_change(e):
            existing = e['Existing Value']
            new = e['New Value']
            if (((existing in [None, '', "None"]) and (new in [None, '', "None"])) or
                (existing == new)):
                return "<i class='fa fa-check' style='color: green'></i> YES"
            return "<i class='fa fa-exclamation-triangle' style='color: red'></i> NO"

        mapped = []
        error_fields = []
        # => Now add the rest of the entries...
        for k, v in modelvars.items():

            v1 = self.field_specific_cleanup(k, v)
            entry = {
                "RFP Param No": varposition.get(k, "50"),
                "RFP Parameter": k,
                "RFP Value": v,
                f"{server} Field Desc": k,
                f"{server} Field Name": k,
                "New Value": v1,
                "Existing Value": opportunity.get(k, ""),
            }

            # Add a check for all fields...
            entry["Match"] = detect_change(entry)
            entry['Post'] = ''
            if (("YES" not in entry['Match']) and
                (entry['New Value'] is not None)):
                entry["Post"] =  f"""\
<input class="form-check-input checkboxinput" type="checkbox" name="post" data-varname="{entry["RFP Parameter"]}" checked=""/>
                """
            mapped.append(entry)

        columns = [
            "RFP Param No", "RFP Parameter", "RFP Value",
            f"{server} Field Desc", f"{server} Field Name"
        ]

        ordercol = 0
        if action == "edit":
            columns.extend(["Existing Value", "New Value", "Match", "Post"])
            # ordercol = 6
        else:
            columns.extend(["New Value"])

        widgets = []
        for name, description, data in [
                ("Mapped", f"Posted to {server}", mapped),
        ]:
            if len(data) == 0:
                continue

            widget = {
                "name": f"{name} Parameters: {plan_name}",
                "description": description,
                "type": "full_width_table_compact_actions",
                "columns": columns,
                "search": True,
                "rows": data,
                "page_size": 50,
                "order": [[ordercol, "asc"]],
                "td_class": "white-space-normal wordwrap",
                "thead_th_class": "",
                "header_components": {
                    "components": [
                        {
                            "template": "action_search"
                        }
                    ]
                }
            }

            widgets.append(widget)

        #=> Add notes..
        notes ="""<ol>\n"""
        if action == "edit":
            notes += f"<li><strong>NOTE: You are UPDATING an existing Opportunity ({opportunity['Name']})</strong></li>"
        else:
            notes += f"<li><strong>NOTE: You are creating/adding a new Opportunity </strong></li>"
        notes += f"""\
<li>Please review existing opportunity, intermediary, and sponsor lists</li>
<li>Some fields require default values such as Closing Date</li>
<li>Note that state, estimated size and other variables will be cleaned to make sure it is consistent with {server} format. Make sure you check the "New Value" column for all variables</li>
"""
        notes += """</ol>"""

        widget = {
            "name": f"Instructions & Notes: {plan_name}",
            "description": "Notes on the post",
            "type": "full_width_text",
            "text": notes,
            "td_class": "white-space-normal wordwrap",
        }
        widgets.insert(0, widget)

        widget = {
            "name": "RFP Workflow",
            "type": "workflow",
            "order": [
                {
                    "name": "Upload RFP",
                    "url": dataseturl
                },
                {
                    "name": "Validate Parameters",
                    "url": validateurl
                },
                {
                    "name": f"Prepare Post to {server}",
                    "url": selecturl,
                },
                f"Approve Post to {server}"
            ],
            "currpos": 3,
            "class": "mb-2"
        }
        widgets.insert(0, widget)

        widget = {
            "name": f"Post to {server}: {plan_name}",
            "description": "Take action on the above content",
            "type": "full_width_form",
            "text": """
<ol>
<li>Check everything</li>
</ol>
            """,
            "hidden_vars": {
                "action": action,
                "opportunityid": action_oppid,
                "checkedvars": ""
            },
            "elements": [
                {
                    "type": "select",
                    "name": "stagename",
                    "id": "stagename",
                    "label": "StageName",
                    "choices": [
                        "Bid",
                        "Onboarding",
                        "Ongoing"
                    ],
                    "selected_choice": "Bid"
                },
                {
                    "type": "textarea",
                    "name": "comments",
                    "id": "comments",
                    "rows": 5,
                    "cols": 120,
                    "placeholder": "Please enter any comments here. They will be saved in the comments field",
                    "value": "",
                    "label": f"Comments",
                },

            ],
            "action": reverse(r.namespace + ":salesforce_dataset_action",
                              kwargs={
                                  'datasetpk': datasetpk
                              }),
            "submit": "Yes",
            "submit_class": "confirmbtn",
            "submit_header": f"Confirm Post to {server}",
            "submit_body": f"By clicking Yes, you are confirming the accuracy of the data. Once you click, the data will be sent to {server}. In case of an update, any prior data will be overwritten in a non reversible manner and for a new record, a duplicate may be created. Please verify on/in {server}.",
            "submit_handler": """
// Extract the checked boxes and post them...
var varlist = new Array();
$('input[type=checkbox]').each(function () {
   if ($(this).prop('checked')) {
       var varname = $(this).data('varname');
       if (typeof varname != 'undefined'){
         varlist.push(varname)
       }
   }
});
$('input[name="checkedvars"]').val(JSON.stringify(varlist));
            """,

            "td_class": "white-space-normal wordwrap",
        }

        widgets.append(widget)


        clean_and_validate_widgets(widgets)

        data = {
            "title": f"Approve Post to {server}: {dataset.name}",
            "sidebar_targets": self.get_sidebar(request, spec),
            "breadcrumbs": [
                {
                    "name": "Datasets",
                    "url": reverse(r.namespace + ":dataset_index")
                },
                {
                    "name": "Validate",
                    "url": validateurl
                },
                {
                    "name": f"Approve {server} Post"
                }
            ],
            "widgets": widgets
        }

        # Make note of any errors you may have seen
        if len(error_fields) > 0:
            messages.warning(request, f"Please double check the following fields: {','.join(error_fields)}")

        template = self.get_template(spec, 'salesforce_dataset_verify')
        return render(request,
                      template,
                      {
                          'app': self,
                          'usecase': usecase,
                          'spec': spec,
                          'basenamespace': r.namespace,
                          'data': data,
                      })

    def salesforce_dataset_action(self, request, spec, datasetpk):
        """
        Post to salesforce
        """

        server = self.salesforce_get_server_name(request, spec)

        action = request.GET.get('action', None)
        action_oppid = request.GET.get('opportunityid', None)
        action_oppid = None if str(action_oppid).lower() == "none" else action_oppid

        # => Get models
        LLMDatasetModel = self.get_model(spec, 'llmdataset')

        # Get the dataset
        try:
            dataset = LLMDatasetModel.objects.get(pk=datasetpk)
        except LLMDatasetModel.DoesNotExist:
            messages.error(request, 'Invalid dataset')
            return HttpResponseRedirect(reverse(r.namespace + ":salesforce_detail",
                                                kwargs={
                                                    "oppid": action_oppid
                                                }))
        dataset_name = dataset.name

        usecase = spec['usecase']
        namespace = spec['namespace']
        r = resolve(request.path)

        # Create an audit record
        if hasattr(self, 'audit_add'):
            self.audit_add(request, spec,
                           action=f"salesforce_{action}",
                           summary=f"Post to {server}",
                           details={
                               'dataset_pk': datasetpk,
                               'dataset_name': dataset_name,
                               'opportunity_id': action_oppid,
                           })

        # What should be redirect to...
        if action == "create":
            messages.info(request, "Created dummy opportunity")
            return HttpResponseRedirect(reverse(r.namespace + ":salesforce_index"))
        else:
            messages.info(request, "Updated dummy opportunity")
            return HttpResponseRedirect(reverse(r.namespace + ":salesforce_detail",
                                                kwargs={
                                                    "oppid": action_oppid
                                                }))

if __name__ == "__main__":

    salesforce = SalesforceClient(cred="acme-salesforce")

    results = salesforce.get_opportunities()
    print(json.dumps(results, indent=4))

    #results = salesforce.describe_opportunity()
    #print(json.dumps(results, indent=4))

